/***
 * Fractal ADL Parser
 * Copyright (C) 2002-2004 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: Eric.Bruneton@rd.francetelecom.com
 *
 * Author: Eric Bruneton
 */

package org.objectweb.fractal.adl.bindings;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.AbstractLoader;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;

/**
 * A {@link org.objectweb.fractal.adl.Loader} to check {@link Binding} nodes in
 * definitions. This loader checks that the from and to interfaces specified in
 * these nodes exist.
 */

public class BindingLoader extends AbstractLoader {

  // --------------------------------------------------------------------------
  // Implementation of the Loader interface
  // --------------------------------------------------------------------------

  public Definition load(final String name, final Map<Object, Object> context)
      throws ADLException {
    final Definition d = clientLoader.load(name, context);
    checkNode(d, context);
    return d;
  }

  // --------------------------------------------------------------------------
  // Checking methods
  // --------------------------------------------------------------------------

  private void checkNode(final Object node, final Map<Object, Object> context)
      throws ADLException {
    if (node instanceof BindingContainer) {
      final Map<String, Interface> itfMap = new HashMap<String, Interface>();
      if (node instanceof InterfaceContainer) {
        for (final Interface itf : ((InterfaceContainer) node).getInterfaces()) {
          itfMap.put("this." + itf.getName(), itf);
        }
      }
      if (node instanceof ComponentContainer) {
        for (final Component comp : ((ComponentContainer) node).getComponents()) {
          if (comp instanceof InterfaceContainer) {
            for (final Interface itf : ((InterfaceContainer) comp)
                .getInterfaces()) {
              itfMap.put(comp.getName() + "." + itf.getName(), itf);
            }
          }
        }
      }
      for (final Binding binding : ((BindingContainer) node).getBindings()) {
        checkBinding(binding, itfMap, context);
      }
      final Set<String> fromItfs = new HashSet<String>();
      for (final Binding binding : ((BindingContainer) node).getBindings()) {
        if (fromItfs.contains(binding.getFrom())) {
          throw new ADLException("Multiple bindings from the same interface",
              binding);
        } else {
          fromItfs.add(binding.getFrom());
        }
      }
    }
    if (node instanceof ComponentContainer) {
      for (final Component comp : ((ComponentContainer) node).getComponents()) {
        checkNode(comp, context);
      }
    }
  }

  protected void checkBinding(final Binding binding,
      final Map<String, Interface> itfs, final Map<Object, Object> context)
      throws ADLException {
    if (binding.getFrom() == null) {
      throw new ADLException("'from' interface missing", binding);
    }
    if (binding.getTo() == null) {
      throw new ADLException("'to' interface missing", binding);
    }
    final Interface fromItf = getInterface(binding.getFrom(), itfs);
    final Interface toItf = getInterface(binding.getTo(), itfs);
    if (fromItf == null && !binding.getFrom().endsWith(".component")
        && !binding.getFrom().endsWith("-controller")) {
      throw new ADLException("No such interface '" + binding.getFrom() + "'",
          binding);
    } else if (toItf == null && !binding.getTo().endsWith(".component")
        && !binding.getTo().endsWith("-controller")) {
      throw new ADLException("No such interface '" + binding.getTo() + "'",
          binding);
    }
    if (fromItf != null && toItf != null) {
      checkBinding(binding, fromItf, toItf, context);
    }
  }

  protected void checkBinding(final Binding binding, final Interface fromItf,
      final Interface toItf, final Map<Object, Object> context)
      throws ADLException {
  }

  protected Interface getInterface(final String name,
      final Map<String, Interface> itfs) {
    return itfs.get(name);
  }
}
